package ch.psi.zmq.imagej;

// EPICS_AD_Viewer.java
// Original authors
// Tim Madden, APS
// Mark Rivers, University of Chicago
import ij.*;
import ij.process.*;

import java.awt.*;

import ij.plugin.*;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.awt.event.*;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Semaphore;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.*;
import javax.swing.Timer;

import org.jeromq.ZMQ;

import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;

public class ZeroMQViewer implements PlugIn {

	private static final Logger logger = Logger.getLogger(ZeroMQViewer.class.getName());

	private ImagePlus img;
	
	private boolean flipX = false;
	private boolean flipY = false;

	private int imageSizeX = 2560;
	private int imageSizeY = 2160;

	// These are used for the frames/second calculation
	private long prevTime;
	private int numImageUpdates;

	private JFrame frame;
	// private JLabel fpsText;

	private volatile boolean isPluginRunning;
	private volatile boolean collect;

	private Timer timer;

	private ZMQ.Context context;
	private ZMQ.Socket socket;

	private JPanel panel_1;
	private JLabel lblNewLabel;
	private JLabel labelFrameRate;
	private JTextField textHostname;
	private JLabel lblHostname;
	private JLabel lblPort;
	private JTextField textPort;
	private JButton btnStart;
	
	private Semaphore semaphore = new Semaphore(1);
	private JLabel lblMethod;
	private JComboBox<String> comboBoxMethod;
	
	private ObjectMapper mapper = new ObjectMapper(new JsonFactory());
	private HeaderInfo hinfo = new HeaderInfo();
	private JLabel lblFlip;
	private JPanel panel;
	private JCheckBox chckbxX;
	private JCheckBox chckbxY;

	public void run(String arg) {
		IJ.showStatus("Running ZeroMQ Viewer");
		try {
			isPluginRunning = true;

			prevTime = System.currentTimeMillis();
			numImageUpdates = 0;
			semaphore.acquire(); // block semaphore
			
			javax.swing.SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					createAndShowGUI();
				}
			});

			
			while (isPluginRunning) {
				semaphore.acquire();
				collect=true;
				hinfo.setVisible(true);
				try{
					if(img!=null){
						img.close();
						img=null;
					}
					String hostname = textHostname.getText();
					int port = Integer.parseInt(textPort.getText());
					String method = (String) comboBoxMethod.getSelectedItem();
					
					
					context = ZMQ.context(1);
					if(method.equals("PULL")){
						socket = context.socket(ZMQ.PULL);
					}
					else if(method.equals("SUB")){
						socket = context.socket(ZMQ.SUB);
					}
					else{
						logger.severe("Method not supported");
						collect=false;
					}
					socket.connect("tcp://"+hostname+":"+port);
					
					logger.info("Connected to: tcp://"+hostname+":"+port);
					
					while(collect){
						byte[] header = socket.recv();
						byte[] content = null;
						if (socket.hasReceiveMore()) {
							content = socket.recv();
						}
						
						readHeader(header);
						readContent(content);
					}
				}
				catch(Exception e){
					logger.log(Level.SEVERE, "",e);
				}
				finally{
					try{
						socket.close();
						context.term();
					}
					catch(Exception e){
					}
				}
			}

			timer.stop();

			img.close();
			frame.setVisible(false);

			IJ.showStatus("Exiting ZeroMQ Viewer");

		} catch (Exception e) {
			logger.log(Level.SEVERE, "", e);
			IJ.showStatus(e.toString());
		}
	}

	@SuppressWarnings("unchecked")
	private void readHeader(byte[] h){
		try{
			String header = new String(h);
//			hinfo.setHeader(header);
			Map<String,Object> m = mapper.readValue(header, new TypeReference<HashMap<String,Object>>(){});
			if(((List<String>) m.get("htype")).contains("array-1.0")){ // currently we only support array-1.0 message types
				List<Integer> shape = (List<Integer>) m.get("shape");
				int nImageSizeX = shape.get(1);
				int nImageSizeY = shape.get(0);
				if(imageSizeX!=nImageSizeX || imageSizeY!=nImageSizeY){
					imageSizeX = nImageSizeX;
					imageSizeY = nImageSizeY;
					
					img.close();
					img=null;
				}
			
				if (img == null) {
					// TODO eventually use ByteProcessor or BinaryProcessor
					// BinaryProcessor p = new ij.process.BinaryProcessor(new
					// ByteProcessor(imageSizeX, imageSizeY));
					img = new ImagePlus("", new ShortProcessor(imageSizeX, imageSizeY));
					img.show();
				}
				img.setTitle(header);
				hinfo.setText(m);
				
			}
			else{
				logger.info("Header type is not supported ...");
				if(img!=null){
					img.close();
					img=null;
				}
			}
		}
		catch(IOException e){
			logger.log(Level.SEVERE, "Unable to parse header", e);
		}
		
//		logger.info(sheader);
	}
	
	private void readContent(byte[] content) {
		try {
			if(content!=null && img!=null){
				// TODO Check whether this is needed
				short[] shorts = new short[content.length / 2];
				ByteBuffer.wrap(content).order(ByteOrder.LITTLE_ENDIAN).asShortBuffer().get(shorts);
				
				ImageProcessor ip = img.getProcessor();
				ip.setPixels(shorts);
				if(flipX){
					ip.flipHorizontal();  
				}
				if(flipY){
					ip.flipVertical();
				}
				
				img.updateAndDraw();
				numImageUpdates++;
			}
		} catch (Exception ex) {
			logger.log(Level.SEVERE, "UpdateImage got exception", ex);
		}
	}

	/**
	 * Create the GUI and show it. For thread safety, this method should be
	 * invoked from the event-dispatching thread.
	 * 
	 * @wbp.parser.entryPoint
	 */
	public void createAndShowGUI() {

		frame = new JFrame("ZeroMQ_Viewer Plugin");
		panel_1 = new JPanel();
		GridBagConstraints gbc_panel_1 = new GridBagConstraints();
		gbc_panel_1.fill = GridBagConstraints.BOTH;
		gbc_panel_1.gridx = 0;
		gbc_panel_1.gridy = 1;
		frame.getContentPane().add(panel_1, BorderLayout.CENTER);
		GridBagLayout gbl_panel_1 = new GridBagLayout();
		gbl_panel_1.columnWidths = new int[] { 0, 0, 0 };
		gbl_panel_1.rowHeights = new int[] { 0, 0, 0, 0, 0, 0, 0 };
		gbl_panel_1.columnWeights = new double[] { 0.0, 1.0, Double.MIN_VALUE };
		gbl_panel_1.rowWeights = new double[] { 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, Double.MIN_VALUE };
		panel_1.setLayout(gbl_panel_1);

		lblHostname = new JLabel("Hostname");
		GridBagConstraints gbc_lblHostname = new GridBagConstraints();
		gbc_lblHostname.insets = new Insets(0, 0, 5, 5);
		gbc_lblHostname.anchor = GridBagConstraints.EAST;
		gbc_lblHostname.gridx = 0;
		gbc_lblHostname.gridy = 0;
		panel_1.add(lblHostname, gbc_lblHostname);

		textHostname = new JTextField();
		GridBagConstraints gbc_textHostname = new GridBagConstraints();
		gbc_textHostname.insets = new Insets(0, 0, 5, 0);
		gbc_textHostname.fill = GridBagConstraints.HORIZONTAL;
		gbc_textHostname.gridx = 1;
		gbc_textHostname.gridy = 0;
		panel_1.add(textHostname, gbc_textHostname);
		textHostname.setColumns(10);

		lblPort = new JLabel("Port");
		GridBagConstraints gbc_lblPort = new GridBagConstraints();
		gbc_lblPort.anchor = GridBagConstraints.EAST;
		gbc_lblPort.insets = new Insets(0, 0, 5, 5);
		gbc_lblPort.gridx = 0;
		gbc_lblPort.gridy = 1;
		panel_1.add(lblPort, gbc_lblPort);

		textPort = new JTextField();
		GridBagConstraints gbc_textPort = new GridBagConstraints();
		gbc_textPort.insets = new Insets(0, 0, 5, 0);
		gbc_textPort.fill = GridBagConstraints.HORIZONTAL;
		gbc_textPort.gridx = 1;
		gbc_textPort.gridy = 1;
		panel_1.add(textPort, gbc_textPort);
		textPort.setColumns(10);
		
		lblMethod = new JLabel("Method");
		GridBagConstraints gbc_lblMethod = new GridBagConstraints();
		gbc_lblMethod.anchor = GridBagConstraints.EAST;
		gbc_lblMethod.insets = new Insets(0, 0, 5, 5);
		gbc_lblMethod.gridx = 0;
		gbc_lblMethod.gridy = 2;
		panel_1.add(lblMethod, gbc_lblMethod);
		
		comboBoxMethod = new JComboBox<String>();
		comboBoxMethod.setModel(new DefaultComboBoxModel<String>(new String[] {"PULL", "SUB"}));
		comboBoxMethod.setSelectedIndex(0);
		GridBagConstraints gbc_comboBoxMethod = new GridBagConstraints();
		gbc_comboBoxMethod.insets = new Insets(0, 0, 5, 0);
		gbc_comboBoxMethod.fill = GridBagConstraints.HORIZONTAL;
		gbc_comboBoxMethod.gridx = 1;
		gbc_comboBoxMethod.gridy = 2;
		panel_1.add(comboBoxMethod, gbc_comboBoxMethod);

		lblNewLabel = new JLabel("Frame/s");
		GridBagConstraints gbc_lblNewLabel = new GridBagConstraints();
		gbc_lblNewLabel.anchor = GridBagConstraints.EAST;
		gbc_lblNewLabel.insets = new Insets(0, 0, 5, 5);
		gbc_lblNewLabel.gridx = 0;
		gbc_lblNewLabel.gridy = 3;
		panel_1.add(lblNewLabel, gbc_lblNewLabel);

		labelFrameRate = new JLabel("0.0");
		GridBagConstraints gbc_labelFrameRate = new GridBagConstraints();
		gbc_labelFrameRate.anchor = GridBagConstraints.EAST;
		gbc_labelFrameRate.insets = new Insets(0, 0, 5, 0);
		gbc_labelFrameRate.gridx = 1;
		gbc_labelFrameRate.gridy = 3;
		panel_1.add(labelFrameRate, gbc_labelFrameRate);

		btnStart = new JButton("Start");
		btnStart.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if(btnStart.getText().equals("Start")){
					// Start data acquisition
					semaphore.release();
					btnStart.setText("Stop");
				}
				else{
					// Stop data acquisition
					collect = false;
					try{
						socket.notifyAll();
					}
					catch(Exception ex){ // This exception can savely be ignored (somewhat most of the time an exception is expected)
					}
					btnStart.setText("Start");
				}
			}
		});
		
		lblFlip = new JLabel("Flip");
		GridBagConstraints gbc_lblFlip = new GridBagConstraints();
		gbc_lblFlip.insets = new Insets(0, 0, 5, 5);
		gbc_lblFlip.gridx = 0;
		gbc_lblFlip.gridy = 4;
		panel_1.add(lblFlip, gbc_lblFlip);
		
		panel = new JPanel();
		GridBagConstraints gbc_panel = new GridBagConstraints();
		gbc_panel.insets = new Insets(0, 0, 5, 0);
		gbc_panel.fill = GridBagConstraints.BOTH;
		gbc_panel.gridx = 1;
		gbc_panel.gridy = 4;
		panel_1.add(panel, gbc_panel);
		
		chckbxX = new JCheckBox("X");
		chckbxX.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				flipX = chckbxX.isSelected();
			}
		});
		panel.add(chckbxX);
		
		chckbxY = new JCheckBox("Y");
		chckbxY.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				flipY = chckbxY.isSelected();
			}
		});
		panel.add(chckbxY);
		
		GridBagConstraints gbc_btnStart = new GridBagConstraints();
		gbc_btnStart.anchor = GridBagConstraints.WEST;
		gbc_btnStart.gridx = 1;
		gbc_btnStart.gridy = 5;
		panel_1.add(btnStart, gbc_btnStart);

		// Display the window.
		frame.pack();
		frame.addWindowListener(new FrameExitListener());
		frame.setVisible(true);

		
		// Update frame rate
		int timerDelay = 2000; // 2 seconds
		timer = new Timer(timerDelay, new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				long time = System.currentTimeMillis();
				double fps = 1000. * numImageUpdates / (double) (time - prevTime);
				labelFrameRate.setText(String.format("%.1f", fps));
				prevTime = time;
				numImageUpdates = 0;
			}
		});
		timer.start();

	}

	public class FrameExitListener extends WindowAdapter {
		public void windowClosing(WindowEvent event) {
			isPluginRunning = false;
		}
	}
}
